//===============================================================================
// Program : psum1.cpp
// Purpose : Sums up elemnts of an array, a[n], using
//           block distribution
//
// Compile : mpicxx -o psum1.x psum1.cpp -O2
//===============================================================================
#include <iostream>
#include <mpi.h>
#include <new>
using namespace std;

// Fuction prototypes............................................................
void para_range(int n1, int n2, int &nprocs, int &irank, int &ista, int &iend);

// Main program starts here......................................................
int main(int argc, char** argv){
  int i;
  int sum;
  int ssum;
  int iproc;
  int nproc;
  int ista;
  int iend;
  int loc_dim;
  int n = 1000;
  int *a;
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&iproc);
  MPI_Comm_size(MPI_COMM_WORLD,&nproc);
  // Call "para_range" to compute lowest and highest iteration ranges for each MPI rank 
  para_range(1,n,nproc,iproc,ista,iend);
  // Generates local a[n] on each MPI process....................................
  loc_dim = iend - ista + 1; // Compute local dimension
  a = new int[loc_dim];      // Allocate memory
  for ( i = 0; i < loc_dim; i++ ){
    a[i] = i + ista;
  }
  // Calculate partial sums for each MPI rank....................................
  sum = 0;
  for ( i = 0; i < loc_dim; i++ ){
    sum = sum + a[i];
  }
  // Calculates total sum........................................................
  MPI_Reduce(&sum,&ssum,1,MPI_INTEGER,MPI_SUM,0,MPI_COMM_WORLD);
  sum = ssum;
  if ( iproc == 0 ){
    cout << "sum = " << sum << endl;
  }
  delete [] a;   // Free memory
  MPI_Finalize();
  return 0;
}

// Functions used my main program................................................
//===============================================================================
// Calculates iteration range and/or array dimension for each MPI process
// Adapted from "RS/6000 SP: Practical MPI Programming", IBM red book
//===============================================================================
void para_range(int n1, int n2, int &nprocs, int &irank, int &ista, int &iend){
  int iwork1;
  int iwork2;
  iwork1 = ( n2 - n1 + 1 )  /  nprocs;
  iwork2 = ( ( n2 - n1 + 1 ) %  nprocs );
  ista = irank * iwork1 + n1 + min(irank, iwork2);
  iend = ista + iwork1 - 1;
  if ( iwork2 > irank ) iend = iend + 1;
}
